#!/usr/bin/env bash

# common.sh - common functions.
# Copyright (c) 2022 ge <https://nixhacks.net/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

log() {
    # Logger. Write message to log_file.
    #
    # Usage: log [-pV] MESSAGE
    #
    # Can read from STDIN. Removes ANSI escape codes from log message.
    # Options:
    # -p    echo message to STDOUT too
    # -V    do not print message to STDOUT in verbose mode

    local message
    local print
    local ignore_verbose

    # Read message from STDIN if args not passed.
    [[ ! -t 0 ]] || [[ "$#" == 0 ]] && message="$(cat <&0)"

    # Set log date format
    log_date_format="${log_date_format:-%d/%b/%Y:%H:%M:%S %z}"

    while (( "$#" )); do
        case "$1" in
            -p) print=1; shift;;
            -V) ignore_verbose=1; shift;;
            -)  message="$(cat <&0)"; shift "$#";;
            *)  message="$*"; shift "$#";;
        esac
    done

    if [ -n "$print" ] || [ -n "$verbose_output" ]; then
        if [ -n "$print" ] || [ -z "$ignore_verbose" ]; then
            echo -e "$message"
        fi
    fi

    echo -e "$message" |
    sed -r 's/\x1B\[(([0-9]+)(;[0-9]+)*)?[m,K,H,f,J]//g' |
    while read -r line; do
        if [ -n "$line" ]; then
            if [ -n "$use_syslog" ]; then
                logger --id="$$" --tag boring_backup -- "$line"
            else
                # shellcheck disable=SC2154
                printf '[%s] %s\n' \
                    "$(date +"$log_date_format")" "$line" >> "$log_file"
            fi
        fi
    done
}

error() {
    # Handle errors
    #
    # Usage: error SCRIPTNAME FUNCNAME LINENO EXIT_CODE COMMAND

    local script lastline funcname lasterr msg
    script="$1"
    funcname="$2"
    lastline="$3"
    lasterr="$4"
    shift 4

    # Display and log message
    msg="ERROR in $script:$lastline: function '$funcname': \
command '$*' failed with exit code $lasterr"

    log -p "\e[31m$msg\nCheck $log_file log for details.\e[0m" >&2

    # Run user defined error handler function if set
    if is_function_set on_error; then
        echo -e "Run \e[1mon_error()\e[0m" | log -p
        on_error "$msg"
    fi

    log -p Exiting with previous errors. >&2
    log -V Backup finished with errors.
    exit 1
}

is_installed() {
    # Test is the program installed.
    # See good answer: https://stackoverflow.com/a/677212
    #
    # Usage: is_installed COMMAND

    if ! command -v "$1" >/dev/null 2>&1; then
        echo "Error: Command $1 not found." \
            "Please install $1 or check your" \
            "PATH if it's actually installed." >&2
        exit 1
    fi
}

is_function_set() {
    # Test is function set or not. Return exit code.
    # Useful with 'if' statement.
    #
    # Usage: is_function_set FUNCTION

    declare -F -- "$1" > /dev/null
    return "$?"
}

remove_if_empty() {
    # Delete empty file
    #
    # Usage: remove_empty FILE

    if [ -f "$1" ]; then
        if [ ! -s "$1" ]; then
            log -p "Backup size is 0 bytes. Removing $1" >&2
            rm "$1"
        fi
    fi
}

set_compression() {
    # Set `cmpr_cmd`, `cmpr_ext` and `tar_ext` by `$1` value.

    # Select compression utility and set filename extension.
    # Refference: https://www.gnu.org/software/tar/manual/html_node/gzip.html
    case "$1" in
    gzip|gz)    cmpr_ext='.gz';    tar_ext='.tar.gz';   cmpr_cmd='gzip';;
    tgz)        cmpr_ext='.gz';    tar_ext='.tgz';      cmpr_cmd='gzip';;
    taz)        cmpr_ext='.gz';    tar_ext='.taz';      cmpr_cmd='gzip';;
    bzip2|bz2)  cmpr_ext='.bz2';   tar_ext='.tar.bz2';  cmpr_cmd='bzip2';;
    tz2)        cmpr_ext='.bz2';   tar_ext='.tz2';      cmpr_cmd='bzip2';;
    tbz2)       cmpr_ext='.bz2';   tar_ext='.tbz2';     cmpr_cmd='bzip2';;
    tbz)        cmpr_ext='.bz2';   tar_ext='.tbz';      cmpr_cmd='bzip2';;
    lzip|lz)    cmpr_ext='.lz';    tar_ext='.tar.lz';   cmpr_cmd='lzip';;
    lzma)       cmpr_ext='.lzma';  tar_ext='.tar.lzma'; cmpr_cmd='lzma';;
    tlz)        cmpr_ext='.lzma';  tar_ext='.tlz';      cmpr_cmd='lzma';;
    lzop|lzo)   cmpr_ext='.lzop';  tar_ext='.tar.lzo';  cmpr_cmd='lzop';;
    xz)         cmpr_ext='.xz';    tar_ext='.tar.xz';   cmpr_cmd='xz';;
    zstd|zst)   cmpr_ext='.zst';   tar_ext='.tar.zst';  cmpr_cmd='zstd --rm';;
    tzst)       cmpr_ext='.zst';   tar_ext='.tzst';     cmpr_cmd='zstd --rm';;
    *)  # No compression
        # shellcheck disable=SC2034
        cmpr_ext=
        # shellcheck disable=SC2034
        tar_ext='.tar'
        # shellcheck disable=SC2034
        cmpr_cmd=
        ;;
    esac

    [ -z "$cmpr_cmd" ] && return # Exit from function if no compression set

    is_installed "${cmpr_cmd%% *}"
}

compress_file() {
    # Compress and echo compressed file name. Remove uncompressed file.
    #
    # Usage: compress_file FILE

    local uncompressed="$1"
    local compressed=

    # Select compression utility and set filename extension.
    if [ -n "$compression" ]; then
        set_compression "$compression"
    else
        cmpr_cmd=
        cmpr_ext=
    fi

    # Compress file
    if [ -n "$cmpr_cmd" ]; then
        log -V "Compressing file $1 by ${cmpr_cmd%% *} ..."
        $cmpr_cmd "$uncompressed" |& log
        compressed="${uncompressed}${cmpr_ext}"

        # Remove original file.
        if [ -f "$compressed" ]; then
            if [ -f "$uncompressed" ]; then
                log "Remove uncompressed file $1"
                rm -- "$uncompressed"
            fi
        fi
    else
        compressed="$uncompressed"
    fi

    echo "$compressed"
}

gen_backup_name() {
    # Generate backup file name. Write resulting string to STDOUT.
    #
    # Usage: gen_backup_name NAME_EXT

    local name
    local name_ext

    name_prefix="${name_prefix:-${backup_script##*/}_}"
    # shellcheck disable=SC2154
    name="$(basename "$path")" # `path` is variable parsed from URI
    name_ext="$1"
    name_date_format="${name_date_format:-_%Y.%m.%d}"
    name_suffix="${name_suffix:--%H%M}"

    date +"${name_prefix}${name}${name_date_format}${name_suffix}${name_ext}"
}
